<?php
// $Id: entity_parser.inc,v 1.3 2009/11/05 20:19:36 pounard Exp $

/**
 * @file
 * Parser and fetcher for Yamm
 */

class EntityParserNoFetcherException extends EntityException {}
class EntityFetcherException extends EntityException {}

/**
 * Get dependencies abstraction
 *
 * TODO implement initial pull
 */
abstract class AbstractEntityFetcher {

  /**
   * Do initial pull
   *
   * @return mixed
   *   Array of entities
   *   FALSE when no entities from server
   */
  public function pull() {
    $entities = $this->_pull();

    if (! is_array($entities)) {
      throw new EntityFetcherException("No result from fetcher");
    }

    return $entities;
  }

  /**
   * Fetch all entity dependencies
   *
   * @param Entity $entity
   * @return array
   *   Array of Entity instances
   */
  public function fetchDependencies(array $dependencies) {
    // Shortcut to avoid fetcher implementation try to fetch empty data
    if (empty($dependencies)) {
      return array();
    }

    $entities = $this->_fetchDependencies($dependencies);

    if (! is_array($entities)) {
      throw new EntityFetcherException("No result from fetcher");
    }

    return $entities;
  }

  /**
   * Override this method with your business stuff
   *
   * @return array
   *   Array of entities
   */
  protected abstract function _pull();

  /**
   * Override this method with your business stuff
   *
   * @param array &$uuid_array
   *   Array of UUID string
   * @return array
   *   Array of entities
   */
  protected abstract function _fetchDependencies(array &$uuid_array);
}

/**
 * XML/RPC custom fetcher
 */
class XmlrpcEntityFetcher extends AbstractEntityFetcher {

  /**
   * Master URL
   * @var string
   */
  protected $_serverUrl = NULL;

  /**
   * Transaction identifier
   * @var string
   */
  protected $_transactionId = NULL;

  /**
   * Main constructor
   *
   * @param string $serverUrl
   *   Master URL
   * @param string $tid
   *   Transaction identifier
   */
  public function __construct($serverUrl, $tid) {
    $this->_serverUrl = $serverUrl;
    $this->_transactionId = $tid;
  }

  /**
   * Get entities from a xmlrpc method call
   *
   * @param string $method
   *   Method to call
   * @param ..
   *   Params to send to server
   * @return unknown_type
   */
  private function &__getEntities($method) {
    $ret = array();

    if (! $this->_serverUrl) {
      throw new EntityFetcherException("No server configured");
    }

    $args = func_get_args();
    $method = array_shift($args);
    array_unshift($args, $this->_transactionId);
    array_unshift($args, $method);
    array_unshift($args, $this->_serverUrl);

    $result = call_user_func_array('yamm_api_xmlrpc_call', $args);

    if ($result === FALSE) {
      throw new EntityFetcherException("Unable to reach server");
    }

    // Result can be empty (no dependencies)
    if (! empty($result['data'])) {
      foreach ($result['data'] as $serializeEntity) {
        $ret[] = Entity::unserialize($serializeEntity);
      }
    }

    unset($result);
    return $ret;
  }

  /**
   * (non-PHPdoc)
   * @see AbstractEntityFetcher#_pull()
   */
  protected function _pull() {
    return $this->__getEntities('yamm.client.pull');
  }

  /**
   * (non-PHPdoc)
   * @see AbstractEntityFetcher#_fetchDependencies($uuid_array)
   */
  protected function _fetchDependencies(array &$uuid_array) {
    return $this->__getEntities('yamm.client.pull.dependencies', $uuid_array);
  }
}

/**
 * Dependecy parser and object saving engine.
 */
class EntityParser {

  /**
   * Fetcher instance
   * @var AbstractEntityFetcher
   */
  protected $_fetcher = NULL;

  /**
   * Main constructor.
   *
   * @param AbstractEntityFetcher $fetcher
   */
  public function __construct(AbstractEntityFetcher $fetcher) {
    $this->_fetcher = $fetcher;
  }

  /**
   * Parse entities, fetch dependencies, and save them
   *
   * @param array $entities
   *   Array of Entity instances
   * @return void
   */
  public function parse() {
    while ($entities = $this->_fetcher->pull()) {
      foreach ($entities as $entity) {
        if (! $this->_alreadyBuilt($entity)) {
          $this->_buildDependencies($entity);
        }
      }
    }
  }

  /**
   * Circular dependencies break
   * @var array
   */
  protected $_built = array();

  /**
   * Check Entity has already been built and mark it built if not
   *
   * @param Entity $entity
   * @return boolean
   */
  protected function _alreadyBuilt(Entity $entity) {
    $uuid = $entity->getUuid();
    // array_key_exists performance is way better than any other method
    if (! array_key_exists($uuid, $this->_built)) {
      $this->_built[$uuid] = TRUE;
      return FALSE;
    }
    else {
      return TRUE;
    }
  }

  /**
   * Remove all entities that already have been built from dependency
   * array.
   */
  protected function _pruneDependencies(array &$uuid_array) {
    foreach ($uuid_array as $key => $uuid) {
      if (array_key_exists($uuid, $this->_built)) {
        unset($uuid_array[$key]);
      }
    }
  }

  /**
   * Build a full grappe of dependencies
   *
   * @param Entity $entity
   * @return void
   */
  protected function _buildDependencies(Entity $entity) {
    $dependencies = $entity->getDependencies();
    $this->_pruneDependencies($dependencies);
    yamm_api_debug("Entity parser got @count dependencies", array('@count' => count($dependencies)));

    // Go and unpack them
    foreach ($this->_fetcher->fetchDependencies($dependencies) as $depEntity) {

      // Check for already builded ones (circular dependencies)
      if (! $this->_alreadyBuilt($depEntity)) {

        // And build it
        $this->_buildDependencies($depEntity);
      }
    }

    $entity->save();
  }
}